!MNH_LIC Copyright 1995-2022 CNRS, Meteo-France and Universite Paul Sabatier
!MNH_LIC This is part of the Meso-NH software governed by the CeCILL-C licence
!MNH_LIC version 1. See LICENSE, CeCILL-C_V1-en.txt and CeCILL-C_V1-fr.txt
!MNH_LIC for details. version 1.
!-----------------------------------------------------------------
!#################
MODULE MODI_ETHETA
!#################
!
INTERFACE
!
#ifndef MNH_OPENACC
FUNCTION ETHETA(KRR,KRRI,PTHLM,PRM,PLOCPEXNM,PATHETA,PSRCM) RESULT(PETHETA)
#else
SUBROUTINE ETHETA(KRR,KRRI,PTHLM,PRM,PLOCPEXNM,PATHETA,PSRCM,PETHETA)
#endif
!
INTEGER                              :: KRR          ! number of moist var.
INTEGER                              :: KRRI         ! number of ice var.
!
REAL, DIMENSION(:,:,:),  INTENT(IN)  ::   PTHLM     ! Conservative pot. temperature
REAL, DIMENSION(:,:,:,:), INTENT(IN) ::   PRM       ! Mixing ratios, where
!                                      PRM(:,:,:,1) = conservative mixing ratio
REAL, DIMENSION(:,:,:),  INTENT(IN)  ::   PLOCPEXNM ! Lv(T)/Cp/Exner at time t-1
REAL, DIMENSION(:,:,:),  INTENT(IN)  ::   PATHETA   ! Atheta
!                                                    
REAL, DIMENSION(:,:,:),  INTENT(IN)  ::   PSRCM     ! Normalized 2dn_order
                                                    ! moment s'r'c/2Sigma_s2
!
#ifndef MNH_OPENACC
REAL,DIMENSION(SIZE(PTHLM,1),SIZE(PTHLM,2),SIZE(PTHLM,3)):: PETHETA ! result
#else
REAL, DIMENSION(:,:,:),  INTENT(OUT):: PETHETA ! result
#endif
!
!
#ifndef MNH_OPENACC
END FUNCTION ETHETA
#else
END SUBROUTINE ETHETA
#endif
!
END INTERFACE
!
END MODULE MODI_ETHETA
!
!   ############################################################################
#ifndef MNH_OPENACC
FUNCTION ETHETA(KRR,KRRI,PTHLM,PRM,PLOCPEXNM,PATHETA,PSRCM) RESULT(PETHETA)
#else
SUBROUTINE ETHETA(KRR,KRRI,PTHLM,PRM,PLOCPEXNM,PATHETA,PSRCM,PETHETA)
#endif
!   ############################################################################
!
!      PURPOSE
!!     -------
!      ETHETA computes the coefficient Etheta in the flottability turbulent
!      flux. This coefficient relates the vertical flux of the virtual potential
!      temperature ( <Thv' W'> ) to the vertical flux of the conservative potential
!      temperature ( <Thl' W'> ).
!
!!**   METHOD
!!     ------
!!
!!     The virtual potential temperature perturbation is linearized in function
!!     of Thl' and Rnp'. The result is
!!        Thv'= ( ZA + ZC * Atheta * 2 * SRC ) Thl' 
!!             +( ZB + ZC * Amoist * 2 * SRC ) Rnp'
!!     From this relation, we can compute the vertical turbulent fluxes.
!!
!!     EXTERNAL
!!     --------
!!
!!        NONE
!!
!!     IMPLICIT ARGUMENTS
!!     ------------------
!!       Module MODD_CST : contains physical constants.
!!         XRV, XRD  : R for water vapor and dry air
!!            
!!     REFERENCE
!!     ---------
!!
!!
!!     AUTHOR
!!     ------
!!       Jean-Marie Carriere      * Meteo-France *
!!
!!     MODIFICATIONS
!!     -------------
!!       Original       20/03/95
!!     
!!       J. Stein       Feb  28, 1996   optimization + Doctorization
!!       J. Stein       Sept 15, 1996   Atheta previously computed
!!       J.-P. Pinty    May  20, 2003   Improve ETHETA expression
!!       M.Moge         April, 2016     Use openACC directives to port the TURB part of Meso-NH on GPU
!!
!!       J.L Redelsperger    03, 2021   Ocean Model Case 
!! ----------------------------------------------------------------------
!
!*       0. DECLARATIONS
!           ------------
USE MODD_CST
USE MODD_DYN_n, ONLY : LOCEAN

#ifdef MNH_OPENACC
USE MODE_MNH_ZWORK, ONLY: MNH_ALLOCATE_FLAT , MNH_RELEASE_FLAT
#endif
use mode_mppdb

IMPLICIT NONE
!
!*       0.1 declarations of arguments and result
!
!
INTEGER                              :: KRR          ! number of moist var.
INTEGER                              :: KRRI         ! number of ice var.
!
REAL, DIMENSION(:,:,:),  INTENT(IN)  ::   PTHLM     ! Conservative pot. temperature
REAL, DIMENSION(:,:,:,:), INTENT(IN) ::   PRM       ! Mixing ratios, where
!                                      PRM(:,:,:,1) = conservative mixing ratio
REAL, DIMENSION(:,:,:),  INTENT(IN)  ::   PLOCPEXNM ! Lv(T)/Cp/Exner at time t-1
REAL, DIMENSION(:,:,:),  INTENT(IN)  ::   PATHETA   ! Atheta
!                                                    
REAL, DIMENSION(:,:,:),  INTENT(IN)  ::   PSRCM     ! Normalized 2dn_order
                                                    ! moment s'r'c/2Sigma_s2
!
#ifndef MNH_OPENACC
REAL,DIMENSION(SIZE(PTHLM,1),SIZE(PTHLM,2),SIZE(PTHLM,3)):: PETHETA ! result
#else
REAL, DIMENSION(:,:,:),  INTENT(OUT):: PETHETA ! result
#endif
!
!
!
!*       0.2 declarations of local variables
!
REAL,DIMENSION(:,:,:), pointer , contiguous :: ZA, ZRW
!                ZA = coeft A, ZRW = total mixing ratio rw
#ifdef MNH_OPENACC
INTEGER  :: IZA, IZRW
#endif
REAL                                  :: ZDELTA  ! = Rv/Rd - 1
INTEGER                               :: JRR     ! moist loop counter
!
!---------------------------------------------------------------------------

!$acc data present( PTHLM, PRM, PLOCPEXNM, PATHETA, PSRCM, PETHETA )

if ( mppdb_initialized ) then
  !Check all in arrays
  call Mppdb_check( pthlm,     "Etheta beg:pthlm"     )
  call Mppdb_check( prm,       "Etheta beg:prm"       )
  call Mppdb_check( plocpexnm, "Etheta beg:plocpexnm" )
  call Mppdb_check( patheta,   "Etheta beg:patheta"   )
  call Mppdb_check( psrcm,     "Etheta beg:psrcm"     )
end if

#ifndef MNH_OPENACC
allocate( za  ( size( pthlm, 1 ), size( pthlm, 2 ), size( pthlm, 3 ) ) )
allocate( zrw ( size( pthlm, 1 ), size( pthlm, 2 ), size( pthlm, 3 ) ) )
#else
iza  = MNH_ALLOCATE_FLAT( za  , size( pthlm, 1 ), size( pthlm, 2 ), size( pthlm, 3 ) )
izrw = MNH_ALLOCATE_FLAT( zrw , size( pthlm, 1 ), size( pthlm, 2 ), size( pthlm, 3 ) )
#endif

!$acc data present( za, zrw )

!
!*       1. COMPUTE ETHETA
!           --------------
!
!
IF (LOCEAN) THEN                                    ! ocean case
!$acc kernels
   PETHETA(:,:,:) =  1.
!$acc end kernels
ELSE
! acc kernels
  IF ( KRR == 0 ) THEN                               ! dry case
!$acc kernels
    PETHETA(:,:,:) = 1.
!$acc end kernels
  ELSE IF ( KRR == 1 ) THEN                           ! only vapor
!$acc kernels
    ZDELTA = (XRV/XRD) - 1.
    PETHETA(:,:,:) = 1. + ZDELTA*PRM(:,:,:,1)
!$acc end kernels
  ELSE                                                ! liquid water & ice present
!$acc kernels
    ZDELTA = (XRV/XRD) - 1.
    ZRW(:,:,:) = PRM(:,:,:,1)
!$acc end kernels

    IF ( KRRI>0 ) THEN  ! rc and ri case
      !$acc kernels
      ZRW(:,:,:) = ZRW(:,:,:) + PRM(:,:,:,3)
      !$acc loop seq
      DO JRR=5,KRR
        ZRW(:,:,:) = ZRW(:,:,:) + PRM(:,:,:,JRR)
      ENDDO
      ZA(:,:,:) = 1. + (                                    &  ! Compute A
                (1.+ZDELTA) * (PRM(:,:,:,1) - PRM(:,:,:,2) - PRM(:,:,:,4)) &
                -ZRW(:,:,:)                                                &
                       )  /  (1. + ZRW(:,:,:))
    !
    !   Etheta = ZA + ZC * Atheta
    !   ZC is computed from line 2 to line 5
    !   - Atheta * 2. * SRC is computed at line 6
    !
      PETHETA(:,:,:) = ZA(:,:,:)                                                 &
          +( PLOCPEXNM(:,:,:) * ZA(:,:,:)                                        &
                 -(1.+ZDELTA) * (PTHLM(:,:,:) + PLOCPEXNM(:,:,:)*(               &
                                                      PRM(:,:,:,2)+PRM(:,:,:,4)))&
                              / (1. + ZRW(:,:,:))                                &
           ) * PATHETA(:,:,:) * 2. * PSRCM(:,:,:)
      !$acc end kernels
    ELSE
      !$acc kernels
      !$acc loop seq
      DO JRR=3,KRR
        ZRW(:,:,:) = ZRW(:,:,:) + PRM(:,:,:,JRR)
      ENDDO
      ZA(:,:,:) = 1. + (                                    &  ! Compute A
                (1.+ZDELTA) * (PRM(:,:,:,1) - PRM(:,:,:,2)) &
                -ZRW(:,:,:)                                 &
                       )  /  (1. + ZRW(:,:,:))
    !
    !   Etheta = ZA + ZC * Atheta
    !   ZC is computed from line 2 to line 5
    !   - Atheta * 2. * SRC is computed at line 6
    !
      PETHETA(:,:,:) = ZA(:,:,:)                                                 &
          +( PLOCPEXNM(:,:,:) * ZA(:,:,:)                                        &
                 -(1.+ZDELTA) * (PTHLM(:,:,:) + PLOCPEXNM(:,:,:)*PRM(:,:,:,2))   &
                              / (1. + ZRW(:,:,:))                                &
           ) * PATHETA(:,:,:) * 2. * PSRCM(:,:,:)
      !$acc end kernels
    END IF
  END IF
! acc end kernels
END IF

if ( mppdb_initialized ) then
  !Check all out arrays
  call Mppdb_check( petheta, "Etheta end:petheta" )
end if

!$acc end data

#ifndef MNH_OPENACC
deallocate (za, zrw)
#else
CALL MNH_RELEASE_FLAT( iza, izrw )
#endif

!$acc end data

!---------------------------------------------------------------------------
!
#ifndef MNH_OPENACC
END FUNCTION ETHETA
#else
END SUBROUTINE ETHETA
#endif
